@@ -53,7 +53,11 @@ from panther_analysis_tool.schemas import (DATA_MODEL_SCHEMA, GLOBAL_SCHEMA,
                                            PACK_SCHEMA, POLICY_SCHEMA,
                                            RULE_SCHEMA, SCHEDULED_QUERY_SCHEMA,
                                            TYPE_SCHEMA)
-from panther_analysis_tool.test_case import DataModel, TestCase
+
+
+from panther_analysis_tool.data_model import DataModel
+from panther_analysis_tool.enriched_event import PantherEvent
+from panther_analysis_tool.log_schemas import user_defined
 
 DATA_MODEL_LOCATION = "./data_models"
 HELPERS_LOCATION = "./global_helpers"
@@ -221,6 +225,20 @@ def datetime_converted(obj: Any) -> Any:
     return obj
 
 
+def _convert_keys_to_lowercase(mapping: Dict[str, Any]) -> Dict[str, Any]:
+    """A helper function for converting top-level dictionary keys to lowercase.
+    Converting keys to lowercase maintains compatibility with how the backend
+    behaves.
+
+    Args:
+        mapping: The dictionary.
+
+    Returns:
+        A new dictionary with each key converted with str.lower()
+    """
+    return {k.lower(): v for k, v in mapping.items()}
+
+
 def zip_analysis(args: argparse.Namespace) -> Tuple[int, str]:
     """Tests, validates, and then archives all policies and rules into a local zip file.
 
@@ -261,7 +279,7 @@ def zip_analysis(args: argparse.Namespace) -> Tuple[int, str]:
                 analysis.append((file_name, f_path, spec))
                 files.add(file_name)
                 files.add("./" + file_name)
-        analysis = filter_analysis(analysis, args.filter)
+        analysis = filter_analysis(analysis, args.filter, args.filter_inverted)
         for analysis_spec_filename, dir_name, analysis_spec in analysis:
             zip_out.write(analysis_spec_filename)
             # datamodels may not have python body
@@ -412,6 +430,39 @@ def update_schemas(args: argparse.Namespace) -> Tuple[int, str]:
     return 0, ""
 
 
+def update_custom_schemas(args: argparse.Namespace) -> Tuple[int, str]:
+    """
+    Updates or creates custom schemas.
+
+    Returns 1 if any file failed to be updated.
+
+    Args:
+        args: The populated Argparse namespace with parsed command-line arguments.
+
+    Returns:
+        A tuple of return code and a placeholder string.
+    """
+    if args.aws_profile is not None:
+        logging.info("Using AWS profile: %s", args.aws_profile)
+        set_env("AWS_PROFILE", args.aws_profile)
+
+    normalized_path = user_defined.normalize_path(args.path)
+    if not normalized_path:
+        return 1, f"path not found: {args.path}"
+
+    uploader = user_defined.Uploader(normalized_path)
+    results = uploader.process()
+    has_errors = False
+    for failed, summary in user_defined.report_summary(normalized_path, results):
+        if failed:
+            has_errors = True
+            logging.error(summary)
+        else:
+            logging.info(summary)
+
+    return int(has_errors), ""
+
+
 def generate_release_assets(args: argparse.Namespace) -> Tuple[int, str]:
     # First, generate the appropriate zip file
     # set the output file to appropriate name for the release: panther-analysis-all.zip
@@ -618,11 +669,15 @@ def test_analysis(args: argparse.Namespace) -> Tuple[int, list]:
         return 1, ["Nothing to test in {}".format(args.path)]
 
     # Apply the filters as needed
+    if getattr(args, "filter_inverted", None) is None:
+        args.filter_inverted = {}
     for key in specs:
-        specs[key] = filter_analysis(specs[key], args.filter)
+        specs[key] = filter_analysis(specs[key], args.filter, args.filter_inverted)
 
     if all((len(specs[key]) == 0 for key in specs)):
-        return 1, ["No analysis in {} matched filters {}".format(args.path, args.filter)]
+        return 1, [
+            f"No analysis in {args.path} matched filters {args.filter} - {args.filter_inverted}"
+        ]
 
     # import each data model, global, policy, or rule and run its tests
     # first import the globals
@@ -636,7 +691,7 @@ def test_analysis(args: argparse.Namespace) -> Tuple[int, list]:
 
     # then, import rules and policies; run tests
     failed_tests, invalid_detection = setup_run_tests(
-        log_type_to_data_model, specs[DETECTION], args.minimum_tests
+        log_type_to_data_model, specs[DETECTION], args.minimum_tests, args.skip_disabled_tests
     )
     invalid_specs.extend(invalid_detection)
 
@@ -665,17 +720,29 @@ def setup_data_models(data_models: List[Any]) -> Tuple[Dict[str, DataModel], Lis
     for analysis_spec_filename, dir_name, analysis_spec in data_models:
         analysis_id = analysis_spec["DataModelID"]
         if analysis_spec["Enabled"]:
-            # load optional python modules
-            module = None
+            body = None
             if "Filename" in analysis_spec:
-                module, load_err = load_module(os.path.join(dir_name, analysis_spec["Filename"]))
+                _, load_err = load_module(os.path.join(dir_name, analysis_spec["Filename"]))
                 # If the module could not be loaded, continue to the next
                 if load_err:
                     invalid_specs.append((analysis_spec_filename, load_err))
                     continue
-                sys.modules[analysis_id] = module
+                data_model_module_path = os.path.join(dir_name, analysis_spec["Filename"])
+                with open(data_model_module_path, 'r') as python_module_file:
+                    body = python_module_file.read()
+
             # setup the mapping lookups
-            data_model = DataModel(analysis_id, analysis_spec["Mappings"], module)
+            params = {
+                'id': analysis_id,
+                'mappings': [_convert_keys_to_lowercase(mapping)
+                             for mapping in analysis_spec["Mappings"]],
+                'versionId': '',
+            }
+
+            if body is not None:
+                params['body'] = body
+
+            data_model = DataModel(params)
             # check if the LogType already has an enabled data model
             for log_type in analysis_spec["LogTypes"]:
                 if log_type in log_type_to_data_model:
@@ -694,11 +761,14 @@ def setup_data_models(data_models: List[Any]) -> Tuple[Dict[str, DataModel], Lis
 
 
 def setup_run_tests(
-    log_type_to_data_model: Dict[str, DataModel], analysis: List[Any], minimum_tests: int
+    log_type_to_data_model: Dict[str, DataModel], analysis: List[Any], minimum_tests: int,
+        skip_disabled_tests: bool
 ) -> Tuple[DefaultDict[str, List[Any]], List[Any]]:
     invalid_specs = []
     failed_tests: DefaultDict[str, list] = defaultdict(list)
     for analysis_spec_filename, dir_name, analysis_spec in analysis:
+        if skip_disabled_tests and not analysis_spec.get("Enabled", False):
+            continue
         analysis_type = analysis_spec["AnalysisType"]
         analysis_id = analysis_spec.get("PolicyID") or analysis_spec["RuleID"]
         print(analysis_id)
@@ -766,7 +836,9 @@ def print_summary(
             print(err_message.format(spec_filename, spec_error))
 
 
-def filter_analysis(analysis: List[Any], filters: Dict[str, List]) -> List[Any]:
+def filter_analysis(
+        analysis: List[Any], filters: Dict[str, List], filters_inverted: Dict[str, List]
+) -> List[Any]:
     if filters is None:
         return analysis
 
@@ -787,6 +859,12 @@ def filter_analysis(analysis: List[Any], filters: Dict[str, List]) -> List[Any]:
             if not set(spec_value).intersection(values):
                 match = False
                 break
+        for key, values in filters_inverted.items():
+            spec_value = analysis_spec.get(key, "")
+            spec_value = spec_value if isinstance(spec_value, list) else [spec_value]
+            if set(spec_value).intersection(values):
+                match = False
+                break
 
         if match:
             filtered_analysis.append((file_name, dir_name, analysis_spec))
@@ -991,7 +1069,7 @@ def _run_tests(
                 test_case = entry
             else:
                 # Set up each test case, including any relevant data models
-                test_case = TestCase(entry, analysis_data_models.get(log_type))
+                test_case = PantherEvent(entry, analysis_data_models.get(log_type))
             if mock_methods:
                 with patch.multiple(analysis_funcs["module"], **mock_methods):
                     result = analysis_funcs["run"](test_case)
@@ -1035,7 +1113,7 @@ def _run_tests(
 
 
 def _verify_test_run(
-    test_case: TestCase,
+    event: PantherEvent,
     mock_methods: Dict[str, MagicMock],
     analysis_funcs: Dict[str, Any],
 ) -> Dict[str, Any]:
@@ -1046,11 +1124,11 @@ def _verify_test_run(
         try:
             if mock_methods:
                 with patch.multiple(analysis_funcs["module"], **mock_methods):
-                    func_out = func(test_case)
+                    func_out = func(event)
                     if not func_out:
                         test_run_result[func_name] = "FAIL"
             else:
-                func_out = func(test_case)
+                func_out = func(event)
                 if not func_out:
                     test_run_result[func_name] = "FAIL"
         except Exception as err:  # pylint: disable=broad-except
@@ -1163,7 +1241,19 @@ def setup_parser() -> argparse.ArgumentParser:
         "required": False,
     }
     skip_test_name = "--skip-tests"
-    skip_test_arg: Dict[str, Any] = {"action": "store_true", "dest": "skip_tests"}
+    skip_test_arg: Dict[str, Any] = {
+        "action": "store_true",
+        "default": False,
+        "dest": "skip_tests",
+        "required": False,
+    }
+    skip_disabled_test_name = "--skip-disabled-tests"
+    skip_disabled_test_arg: Dict[str, Any] = {
+        "action": "store_true",
+        "default": False,
+        "dest": "skip_disabled_tests",
+        "required": False,
+    }
     ignore_extra_keys_name = "--ignore-extra-keys"
     ignore_extra_keys_arg: Dict[str, Any] = {
         "required": False,
@@ -1177,7 +1267,7 @@ def setup_parser() -> argparse.ArgumentParser:
         + "managing Panther policies and rules.",
         prog="panther_analysis_tool",
     )
-    parser.add_argument("--version", action="version", version="panther_analysis_tool 0.6.1")
+    parser.add_argument("--version", action="version", version="panther_analysis_tool 0.7.1")
     parser.add_argument("--debug", action="store_true", dest="debug")
     subparsers = parser.add_subparsers()
 
@@ -1194,6 +1284,7 @@ def setup_parser() -> argparse.ArgumentParser:
     release_parser.add_argument(out_name, **out_arg)
     release_parser.add_argument(path_name, **path_arg)
     release_parser.add_argument(skip_test_name, **skip_test_arg)
+    release_parser.add_argument(skip_disabled_test_name, **skip_disabled_test_arg)
     release_parser.set_defaults(func=generate_release_assets)
 
     test_parser = subparsers.add_parser(
@@ -1203,6 +1294,7 @@ def setup_parser() -> argparse.ArgumentParser:
     test_parser.add_argument(min_test_name, **min_test_arg)
     test_parser.add_argument(path_name, **path_arg)
     test_parser.add_argument(ignore_extra_keys_name, **ignore_extra_keys_arg)
+    test_parser.add_argument(skip_disabled_test_name, **skip_disabled_test_arg)
     test_parser.set_defaults(func=test_analysis)
 
     publish_parser = subparsers.add_parser(
@@ -1247,6 +1339,7 @@ def setup_parser() -> argparse.ArgumentParser:
     publish_parser.add_argument(min_test_name, **min_test_arg)
     publish_parser.add_argument(out_name, **out_arg)
     publish_parser.add_argument(skip_test_name, **skip_test_arg)
+    publish_parser.add_argument(skip_disabled_test_name, **skip_disabled_test_arg)
     publish_parser.set_defaults(func=publish_release)
 
     upload_parser = subparsers.add_parser(
@@ -1258,14 +1351,15 @@ def setup_parser() -> argparse.ArgumentParser:
     upload_parser.add_argument(out_name, **out_arg)
     upload_parser.add_argument(path_name, **path_arg)
     upload_parser.add_argument(skip_test_name, **skip_test_arg)
+    upload_parser.add_argument(skip_disabled_test_name, **skip_disabled_test_arg)
     upload_parser.add_argument(ignore_extra_keys_name, **ignore_extra_keys_arg)
     upload_parser.set_defaults(func=upload_analysis)
 
-    update_schemas_parser = subparsers.add_parser(
+    update_managed_schemas_parser = subparsers.add_parser(
         "update-schemas", help="Update managed schemas on a Panther deployment."
     )
-    update_schemas_parser.add_argument(aws_profile_name, **aws_profile_arg)
-    update_schemas_parser.set_defaults(func=update_schemas)
+    update_managed_schemas_parser.add_argument(aws_profile_name, **aws_profile_arg)
+    update_managed_schemas_parser.set_defaults(func=update_schemas)
 
     zip_parser = subparsers.add_parser(
         "zip", help="Create an archive of local policies and rules for uploading to Panther."
@@ -1275,6 +1369,7 @@ def setup_parser() -> argparse.ArgumentParser:
     zip_parser.add_argument(out_name, **out_arg)
     zip_parser.add_argument(path_name, **path_arg)
     zip_parser.add_argument(skip_test_name, **skip_test_arg)
+    zip_parser.add_argument(skip_disabled_test_name, **skip_disabled_test_arg)
     zip_parser.set_defaults(func=zip_analysis)
 
     zip_schemas_parser = subparsers.add_parser(
@@ -1286,6 +1381,16 @@ def setup_parser() -> argparse.ArgumentParser:
     zip_schemas_parser.add_argument(out_name, **out_arg)
     zip_schemas_parser.set_defaults(func=zip_managed_schemas)
 
+    update_custom_schemas_parser = subparsers.add_parser(
+        "update-custom-schemas",
+        help="Update or create custom schemas on a Panther deployment."
+    )
+    update_custom_schemas_parser.add_argument(aws_profile_name, **aws_profile_arg)
+    custom_schemas_path_arg = path_arg.copy()
+    custom_schemas_path_arg['help'] = "The relative or absolute path to Panther custom schemas."
+    update_custom_schemas_parser.add_argument(path_name, **custom_schemas_path_arg)
+    update_custom_schemas_parser.set_defaults(func=update_custom_schemas)
+
     return parser
 
 
@@ -1359,13 +1464,18 @@ def zip_managed_schemas(args: argparse.Namespace) -> Tuple[int, str]:
 
 
 # Parses the filters, expects a list of strings
-def parse_filter(filters: List[str]) -> Dict[str, Any]:
+def parse_filter(filters: List[str]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
     parsed_filters = {}
+    parsed_filters_inverted = {}
     for filt in filters:
         split = filt.split("=")
         if len(split) != 2 or split[0] == "" or split[1] == "":
             logging.warning("Filter %s is not in format KEY=VALUE, skipping", filt)
             continue
+        # Check for "!="
+        invert_filter = split[0].endswith("!")
+        if invert_filter:
+            split[0] = split[0][:-1]  # Remove the trailing "!"
         key = split[0]
         if not any(
             (
@@ -1380,8 +1490,11 @@ def parse_filter(filters: List[str]) -> Dict[str, Any]:
         ):
             logging.warning("Filter key %s is not a valid filter field, skipping", key)
             continue
-        parsed_filters[key] = split[1].split(",")
-    return parsed_filters
+        if invert_filter:
+            parsed_filters_inverted[key] = split[1].split(",")
+        else:
+            parsed_filters[key] = split[1].split(",")
+    return parsed_filters, parsed_filters_inverted
 
 
 def set_env(key: str, value: str) -> None:
@@ -1399,7 +1512,9 @@ def run() -> None:
     )
 
     if getattr(args, "filter", None) is not None:
-        args.filter = parse_filter(args.filter)
+        args.filter, args.filter_inverted = parse_filter(args.filter)
+    if getattr(args, "filter_inverted", None) is None:
+        args.filter_inverted = {}
 
     # Although not best practice, the alternative is ugly and significantly harder to maintain.
     if bool(getattr(args, "ignore_extra_keys", None)):
@@ -1411,6 +1526,7 @@ def run() -> None:
     except Exception as err:  # pylint: disable=broad-except
         # Catch arbitrary exceptions without printing help message
         logging.warning('Unhandled exception: "%s"', err)
+        logging.debug('Full error traceback:', exc_info=err)
         sys.exit(1)
 
     if return_code == 1:
